from itertools import tee
from typing import Dict, Iterator, List, Sequence, Tuple

from brown_clustering.defaultvaluedict import DefaultValueDict

Corpus = Sequence[Sequence[str]]


class BigramCorpus:
    def __init__(
            self,
            corpus: Corpus,
            alpha: float = 1,
            start_symbol: str = '<s>',
            end_symbol: str = '</s>',
            min_count: int = 0
    ):
        self.vocabulary: Dict[str, int] = DefaultValueDict(0)

        self.gather_vocab(corpus, min_count)

        word_count = len(self.vocabulary) + 2
        self.alpha = alpha
        self.n = alpha * word_count * word_count
        self.unigrams: Dict[str, float] = DefaultValueDict(alpha * word_count)
        self.bigrams: Dict[Tuple[str, str], float] = DefaultValueDict(alpha)
        self.gather_statistics(corpus, start_symbol, end_symbol)

    def gather_vocab(self, corpus: Corpus, min_count: int):
        for sentence in corpus:
            for word in sentence:
                self.vocabulary[word] += 1

        self.vocabulary = dict(filter(
            lambda x: x[1] >= min_count,
            self.vocabulary.items()
        ))

    def gather_statistics(
            self,
            corpus: Corpus,
            start_symbol: str = '<s>',
            end_symbol: str = '</s>',
    ):
        for sentence in corpus:
            act_sentence = [start_symbol] + [
                w for w in sentence if w in self.vocabulary
            ] + [end_symbol]

            for word in act_sentence:
                self.unigrams[word] += 1

            grams = two_grams(act_sentence)
            for w1, w2 in grams:
                self.n += 1
                self.bigrams[(w1, w2)] += 1

    def bigram_propa(
            self,
            cluster1: Sequence[str],
            cluster2: Sequence[str]
    ) -> float:
        return sum(
            self.bigrams[(w1, w2)]
            for w1 in cluster1
            for w2 in cluster2
        ) / self.n

    def unigram_propa(self, cluster: Sequence[str]) -> float:
        return sum(
            self.unigrams[w]
            for w in cluster
        ) / self.n

    def ranks(self) -> List[Tuple[str, int]]:
        return sorted(self.vocabulary.items(), key=lambda x: (-x[1], x[0]))

    def print_stats(self):
        extended_vocab = len(self.vocabulary) + 2
        alpha_bonus = self.alpha * extended_vocab * extended_vocab

        print(f"Vocab count: {len(self.vocabulary)}")
        print(f"Token count: {sum(self.vocabulary.values())}")
        print(f"unique 2gram count: {len(self.bigrams)}")
        print(f"2gram count: {self.n - alpha_bonus}")
        print(f"Laplace smoothing: {self.alpha}")


def two_grams(sequence: Sequence) -> Iterator[Tuple]:
    iterables = tee(sequence, 2)
    next(iterables[1], None)
    return zip(*iterables)



from copy import deepcopy
from typing import List

import numpy as np
from numba import jit, prange

from brown_clustering.data import BigramCorpus


@jit(nopython=True, parallel=True)
def _q_l(used, p1, p2, q2, x):
    n = p1.shape[0]
    px = p1[x]

    for i in prange(n):
        if not used[i]:
            continue
        pxc = p2[x, i]
        pc = p1[i]
        q2[x, i] = pxc * np.log(pxc / (pc * px))


@jit(nopython=True, parallel=True)
def _q_r(used, p1, p2, q2, x):
    n = p1.shape[0]
    px = p1[x]

    for i in prange(n):
        if not used[i]:
            continue
        pxc = p2[i, x]
        pc = p1[i]
        q2[i, x] = pxc * np.log(pxc / (pc * px))


@jit(nopython=True, parallel=True)
def diag_l2(used, l2):
    n = l2.shape[0]
    for i in prange(n):
        for j in prange(n):
            if not used[i] or not used[j] or i >= j:
                l2[i, j] = -np.inf


@jit(nopython=True, parallel=True)
def _q_l_v(used, l2, p1, p2, x):
    n = p2.shape[0]
    px = p1[x]
    for i in prange(n):
        if i == x or not used[i]:
            continue
        for j in prange(n):
            if j == x or not used[j]:
                continue
            pc = p1[i] + p1[j]
            pcx = p2[i, x] + p2[j, x]
            l2[i, j] += pcx * np.log(pcx / (pc * px))


@jit(nopython=True, parallel=True)
def _q_r_v(used, l2, p1, p2, x):
    n = p2.shape[0]
    px = p1[x]
    for i in prange(n):
        if i == x or not used[i]:
            continue
        for j in prange(n):
            if j == x or not used[j]:
                continue
            pc = p1[i] + p1[j]
            pcx = p2[x, i] + p2[x, j]
            l2[i, j] += pcx * np.log(pcx / (pc * px))


@jit(nopython=True, parallel=True)
def _q_l_n(used, l2, p1, p2, x):
    n = p2.shape[0]
    px = p1[x]
    for i in prange(n):
        if i == x or not used[i]:
            continue
        for j in prange(n):
            if j == x or not used[j]:
                continue
            pc = p1[i] + p1[j]
            pcx = p2[i, x] + p2[j, x]
            l2[i, j] -= pcx * np.log(pcx / (pc * px))


@jit(nopython=True, parallel=True)
def _q_r_n(used, l2, p1, p2, x):
    n = p2.shape[0]
    px = p1[x]
    for i in prange(n):
        if i == x or not used[i]:
            continue
        for j in prange(n):
            if j == x or not used[j]:
                continue
            pc = p1[i] + p1[j]
            pcx = p2[x, i] + p2[x, j]
            l2[i, j] -= pcx * np.log(pcx / (pc * px))


@jit(nopython=True, parallel=True)
def _delta_v(used, p1, p2, q2, x):
    n = p1.shape[0]
    ret = np.zeros_like(p1)

    for i in prange(n):
        if not used[i]:
            continue
        pij = p2[i, i] + p2[i, x] + p2[x, i] + p2[x, x]
        pi = pj = p1[x] + p1[i]
        ret[i] += pij * np.log(pij / (pi * pj))
        ret[i] -= q2[i, i]
        ret[i] -= q2[i, x]
        ret[i] -= q2[x, i]
        ret[i] -= q2[x, x]

        ppi = p1[i] + p1[x]
        for j in prange(n):
            if j == i or j == x or not used[j]:
                continue

            ret[i] -= q2[i, j]
            ret[i] -= q2[j, i]
            ret[i] -= q2[x, j]
            ret[i] -= q2[j, x]

            ppij = p2[i, j] + p2[x, j]
            ppji = p2[j, i] + p2[j, x]

            ppj = p1[j]
            nom = 1 / (ppi * ppj)
            ret[i] += ppji * np.log(ppji * nom)
            ret[i] += ppij * np.log(ppij * nom)

    return ret


@jit(nopython=True, parallel=True)
def _update_delta(mask, l2, p1, p2, q2, x):
    _q_l_v(mask, l2, p1, p2, x)
    _q_r_v(mask, l2, p1, p2, x)

    n = l2.shape[0]

    for i in prange(n):
        if not mask[i]:
            continue
        for j in prange(n):
            if not mask[j]:
                continue
            l2[i, j] -= (
                    + q2[i, x]
                    + q2[j, x]
                    + q2[x, i]
                    + q2[x, j]
            )


@jit(nopython=True, parallel=True)
def _reduce_delta(used, l2, p1, p2, q2, x):
    _q_l_n(used, l2, p1, p2, x)
    _q_r_n(used, l2, p1, p2, x)

    n = l2.shape[0]
    for i in prange(n):
        if not used[i]:
            continue
        for j in prange(n):
            if not used[j]:
                continue
            l2[i, j] += (
                    + q2[i, x]
                    + q2[j, x]
                    + q2[x, i]
                    + q2[x, j]
            )


@jit(nopython=True, parallel=True)
def _update_heuristic(used, l2, p1, p2, q2, x):
    _q_l(used, p1, p2, q2, x)
    _q_r(used, p1, p2, q2, x)

    _update_delta(used, l2, p1, p2, q2, x)
    deltas = _delta_v(used, p1, p2, q2, x)
    l2[:, x] = deltas
    l2[x, :] = deltas
    diag_l2(used, l2)


@jit(nopython=True)
def _combine_clusters(used, l2, p1, p2, q2, i, j):
    n = p2.shape[0]
    _reduce_delta(used, l2, p1, p2, q2, i)
    _reduce_delta(used, l2, p1, p2, q2, j)
    p1[i] += p1[j]
    for k in prange(n):
        p2[i, k] += p2[j, k]
    for k in prange(n):
        p2[k, i] += p2[k, j]


class ClusteringHelper:
    def __init__(self, corpus: BigramCorpus, max_words: int):
        self.m = 0
        self.clusters: List[List[str]] = [[] for _ in range(max_words)]
        self.p1 = np.zeros(max_words, dtype=float)
        self.p2 = np.zeros((max_words, max_words), dtype=float)
        self.q2 = np.zeros((max_words, max_words), dtype=float)
        self.l2 = np.zeros((max_words, max_words), dtype=float)
        self.used = np.zeros(max_words, dtype=bool)
        self.max_words = max_words
        self.corpus = corpus

    def copy_clusters(self) -> List[List[str]]:
        return [
            c
            for c, used in zip(deepcopy(self.clusters), self.used)
            if used
        ]

    def append_cluster(self, words: List[str]):
        new_i = self.used.argmin()
        self.clusters[new_i] = words

        self.p1[new_i] = self.corpus.unigram_propa(words)

        for i in range(self.max_words):
            self.p2[new_i, i] = self.corpus.bigram_propa(
                words,
                self.clusters[i]
            )
            self.p2[i, new_i] = self.corpus.bigram_propa(
                self.clusters[i],
                words
            )
        self.p2[new_i, new_i] = self.corpus.bigram_propa(words, words)
        self.used[new_i] = True
        _update_heuristic(self.used, self.l2, self.p1, self.p2, self.q2, new_i)

        self.m += 1

    def merge_clusters(self, i: int, j: int):
        assert self.used[i] and self.used[j]
        self.clusters[i].extend(self.clusters[j])
        self.clusters[j] = []
        self.m -= 1

        _combine_clusters(self.used, self.l2, self.p1, self.p2, self.q2, i, j)
        self.used[j] = False

        _update_heuristic(self.used, self.l2, self.p1, self.p2, self.q2, i)



import heapq
from collections import defaultdict, deque
from typing import Deque, Dict, List, Sequence, Tuple

import numpy as np
from tqdm import tqdm

from brown_clustering.data import BigramCorpus
from brown_clustering.helper import ClusteringHelper


class BrownClustering:
    def __init__(self, corpus: BigramCorpus, m: int):
        self.m = m
        self.corpus = corpus
        self.vocabulary = corpus.vocabulary
        self.helper = ClusteringHelper(corpus, self.m + 1)
        self._codes: Dict[str, Deque[str]] = defaultdict(lambda: deque())

    def codes(self) -> Dict[str, str]:
        return {
            key: ''.join(value)
            for key, value in self._codes.items()
        }

    def merge_best(self) -> Tuple[int, int]:
        benefit = self.helper.l2
        i, j = np.unravel_index(benefit.argmax(), benefit.shape)

        for word in self.helper.clusters[i]:
            self._codes[word].appendleft("0")

        for word in self.helper.clusters[j]:
            self._codes[word].appendleft("1")
        self.helper.merge_clusters(i, j)
        return i, j

    def _code_similarity(
            self, code1: Sequence[str], code2: Sequence[str]
    ) -> int:
        count = 0
        for w1, w2 in zip(code1, code2):
            if w1 == w2:
                count += 1
            else:
                return count
        return count

    def get_similar(self, word: str, cap: int = 10) -> List[Tuple[str, str]]:
        tmp = self.codes()
        if word not in tmp:
            return []
        code = tmp[word]
        del tmp[word]

        best = heapq.nlargest(
            iterable=tmp.items(),
            n=cap,
            key=lambda it: self._code_similarity(code, it[1])
        )
        return best

    def train(self) -> List[List[str]]:
        words = self.corpus.ranks()

        for w, count in tqdm(words):
            self.helper.append_cluster([w])
            if self.helper.m > self.m:
                self.merge_best()

        clusters = self.helper.copy_clusters()

        for _ in range(len(clusters) - 1):
            self.merge_best()

        return clusters



from pathlib import Path

from setuptools import find_packages, setup

base_path = Path(__file__).parent
long_description = (base_path / "README.md").read_text()

setup(
    name='brown_clustering',
    packages=find_packages(include=("brown_clustering",)),
    version='0.1.4',
    description='Fast Brown Clustering',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='bfuchs',
    author_email="benedikt.fuchs.staw@hotmail.com",
    license="MIT",
    python_requires=">=3.6",
    install_requires=[
        "numpy<1.21",
        "numba",
        "tqdm",
    ],
    extras_require={
        "dev": [
            "pytest",
            "pytest-mypy",
            "pytest-profiling",
            "pytest-flake8",
            "flake8<4.0.0",
            "flake8-isort",
        ]
    },
)



from typing import Generic, TypeVar

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class DefaultValueDict(dict, Generic[_KT, _VT]):
    def __init__(self, default_value: _VT, *args, **kwargs):
        self.default_value = default_value
        super().__init__(*args, **kwargs)

    def __getitem__(self, item: _KT):
        return self.get(item, self.default_value)
